package R;

import calculatecv.Config;
import calculatecv.ExpObs;
import calculatecv.Go;
import calculatecv.GoThread;
import calculatecv.LogEventType;
import calculatecv.Output;
import calculatecv.Step3Panel;
import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Scanner;
import rcaller.RCaller;
import rcaller.RCode;

/**
 * Helper methods for calling functions from the R script through Java
 * 
 * @author Robert Eisinger
 */
public class R {

    private static RCaller caller;
    private static RCode code;
    private static ArrayList<ExpObs> expObs; //for each subject
    private static String allCode;
    
    //maps from subject name to a hashmap that maps from model to a list {alpha before outliers, beta before outliers, sigmaint before outliers}
    public static HashMap<Integer, HashMap<String, String[]>> plotData = new HashMap<Integer, HashMap<String, String[]>>();

    /**
     * Initializes the list of Expected/Observed for each subject. All the R
     * functions use this list, and so it is critical to load this list 
     * before using it in subsequent R methods
     */
    public static void init() {
        expObs = Config.input.getExpectedObserved();
    }

    /**
     * Cleans the current RCaller and sets up a new run by loading 
     * the R file and adding it to the RCaller's RCode object
     */
    public static void clean() {
        caller = new RCaller();
        code = new RCode();
        caller.setRscriptExecutable(Config.RScript);

        caller.cleanRCode();
        if(allCode == null)
            allCode = readFile("depend/PsiMLE_JavaVersion.R");
        
        code.addRCode(allCode);
    }

    /**
     * Runs the discontinuous model for each subject and returns the estimated parameters.
     * @param splitAt the value for the point of discontinuity
     * @return a map from parameter name to a list of each subject's parameter value
     */
    public static LinkedHashMap<String, ArrayList<String>> runDiscontinuousModel(double splitAt) {

        LinkedHashMap<String, ArrayList<String>> out = new LinkedHashMap<String, ArrayList<String>>();

        for (int i = 0; i < expObs.size(); i++) {
            if(!GoThread.run) {
                return null;
            }
            
            clean();

            Go.loadingWindow.addTextToLog("Fitting parameters for " + expObs.get(i).getName() + "...", true);

            boolean failed = false;
            try {
                double[] exp = expObs.get(i).getExp();
                double[] obs = expObs.get(i).getObs();
                code.addDoubleArray("exp" + i, exp);
                code.addDoubleArray("obs" + i, obs);

                String call = "out" + i + " <- " + RConfig.discontinuousModel + "(exp" + i + ",obs" + i + ",discontPoint="+splitAt+");";
                code.addRCode(call);

                caller.setRCode(code);
                caller.runAndReturnResult("out" + i);
            } catch (Exception e) {
                failed = true;
            }
            
            if(extractFromModel("logLik").equals("-Inf") || extractFromModel("logLik").equals("Inf")) {
                failed = true;
            }
            

            String[] params = {"alpha","betaA","betaB","sigmaA","sigmaB","alpha_ci-","alpha_ci+","betaA_ci-","betaA_ci+","betaB_ci-","betaB_ci+","sigmaA_ci-","sigmaA_ci+","sigmaB_ci-","sigmaB_ci+"};
      
            if (failed) {
                for(String param : params) {
                    addToHashMap(out,param,"NA");
                }

                Go.loadingWindow.addTextToLog("FAILED\nUnable to fit parameters.", LogEventType.FAILURE);

            } else {
                addToHashMap(out, "alpha", extractFromModel("alpha"));
                addToHashMap(out, "betaA", extractFromModel("betaA"));
                addToHashMap(out, "betaB", extractFromModel("betaB"));
                addToHashMap(out, "sigmaA", extractFromModel("sigmaA"));
                addToHashMap(out, "sigmaB", extractFromModel("sigmaB"));
                addToHashMap(out, "alpha_ci-", calculateCILower(extractFromModel("alpha"),extractFromModel("alpha_se")));
                addToHashMap(out, "alpha_ci+", calculateCIUpper(extractFromModel("alpha"),extractFromModel("alpha_se")));
                addToHashMap(out, "betaA_ci-", calculateCILower(extractFromModel("betaA"),extractFromModel("betaA_se")));
                addToHashMap(out, "betaA_ci+", calculateCIUpper(extractFromModel("betaA"),extractFromModel("betaA_se")));
                addToHashMap(out, "betaB_ci-", calculateCILower(extractFromModel("betaB"),extractFromModel("betaB_se")));
                addToHashMap(out, "betaB_ci+", calculateCIUpper(extractFromModel("betaB"),extractFromModel("betaB_se")));
                addToHashMap(out, "sigmaA_ci-", calculateCILower(extractFromModel("sigmaA"),extractFromModel("sigmaA_se")));
                addToHashMap(out, "sigmaA_ci+", calculateCIUpper(extractFromModel("sigmaA"),extractFromModel("sigmaA_se")));
                addToHashMap(out, "sigmaB_ci-", calculateCILower(extractFromModel("sigmaB"),extractFromModel("sigmaB_se")));
                addToHashMap(out, "sigmaB_ci+", calculateCIUpper(extractFromModel("sigmaB"),extractFromModel("sigmaB_se")));
                
                
                //For plotting the data
                HashMap<String, String[]> entry = plotData.get(i);
                if(entry == null) {
                    entry = new HashMap<String, String[]>();
                }
                entry.put(RConfig.discontinuousModel,new String[]{extractFromModel("alpha"),extractFromModel("betaA"),extractFromModel("betaB"),extractFromModel("sigmaA"),extractFromModel("sigmaB"),Double.toString(splitAt)});
                plotData.put(i, entry);

                Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
            }

        }

        return out;

    }

    
    public static LinkedHashMap<String, ArrayList<String>> runBayesModel(String s) {
        LinkedHashMap<String, ArrayList<String>> out = new LinkedHashMap<String, ArrayList<String>>();
        for(int i = 0; i < expObs.size(); i++) {
            if(!GoThread.run) 
                return null;
            
            clean();
            
            Go.loadingWindow.addTextToLog("Estimating parameters for " + expObs.get(i).getName() + "...", true);
            
            boolean failed = false;
            try {
                double[] exp = expObs.get(i).getExp();
                double[] obs = expObs.get(i).getObs();
                code.addDoubleArray("exp" + i, exp);
                code.addDoubleArray("obs" + i, obs);
                
                boolean outlier = false;
                String outlierText = "remove.outliers=FALSE";
                if (Config.step3Panel.outlierIsSelected()) {
                    outlier = true;
                    outlierText = "remove.outliers=TRUE,outliers.sd="+Double.parseDouble(Config.step3Panel.outliersSD());
                }

                String modelstring = "bayes.generateStringModel(\"" + s + "\"," + 
                               Config.baysPanel.getAlphaInfo() + "," + Config.baysPanel.getBetaInfo() + "," + Config.baysPanel.getSigmaInfo() + ")";
                
                String plotting = "plotting=FALSE,filename=\"\"";
                if(Config.baysPanel.plotDropBoxSelected()) {
                    plotting = "plotting=TRUE,filename=" + "\""+
                            (Config.step4Panel.getOutputDir()+Config.step4Panel.getSavePlotImagesPrefix()+"_Bayes_"+expObs.get(i).getName()).replaceAll("\\\\","\\\\\\\\")+
                            ".pdf\")";
                }
                
                String call2 = "bayesmodel" + i + " <- bayes.estimate(model.name=\""+s+"\",intensity=exp"+i+",response=obs"+i+",string.model="
                        +modelstring+",n.adapt=100, n.burn=1000, n.chains=4, n.saved=4000,"+
                        outlierText+","+plotting+    ")";
                String finall = "invisible(capture.output("+call2+"))";
                code.addRCode(finall);
                code.addRCode("bayesmodel"+i+"$string.model=\"deleted\"");
                
                caller.setRCode(code);
                System.err.println(finall);
                
                caller.runAndReturnResult("bayesmodel"+i);
                
                System.out.println(caller.getParser().getNames());
                
                
                
                if (failed) {
                    addToHashMap(out, "subject", expObs.get(i).getName());
                    addToHashMap(out, "alpha", "NA");
                    addToHashMap(out, "beta", "NA");
                    addToHashMap(out, "sigma", "NA");
                    addToHashMap(out, "alpha_ci-", "NA");
                    addToHashMap(out, "alpha_ci+", "NA");
                    addToHashMap(out, "beta_ci-", "NA");
                    addToHashMap(out, "beta_ci+", "NA");
                    addToHashMap(out, "sigma_ci-", "NA");
                    addToHashMap(out, "sigma_ci+", "NA");


                    if (outlier) {
                        addToHashMap(out, "num_outliers", "NA");
                    }

                    Go.loadingWindow.addTextToLog("FAILED\nUnable to fit parameters.", LogEventType.FAILURE);

                } else {
                    addToHashMap(out, "subject", expObs.get(i).getName());
                    addToHashMap(out, "alpha", extractFromModel("alpha"));
                    addToHashMap(out, "beta", extractFromModel("beta"));
                    addToHashMap(out, "sigma", extractFromModel("sigma"));
                    addToHashMap(out, "alpha_ci-",extractFromModel("alpha_ciminus"));
                    addToHashMap(out, "alpha_ci+", extractFromModel("alpha_ciplus"));
                    addToHashMap(out, "beta_ci-", extractFromModel("beta_ciminus"));
                    addToHashMap(out, "beta_ci+", extractFromModel("beta_ciplus"));
                    addToHashMap(out, "sigma_ci-", extractFromModel("sigma_ciminus"));
                    addToHashMap(out, "sigma_ci+", extractFromModel("sigma_ciplus"));
                    
                    
                    if (outlier) {
                        addToHashMap(out, "num_outliers", extractFromModel("num_outliers"));
                    }
                
                   /* HashMap<String, String[]> entry = plotData.get(i);
                    if(entry == null) {
                        entry = new HashMap<String, String[]>();
                    }

                    entry.put(s,new String[]{extractFromModel("alpha"),extractFromModel("beta"),extractFromModel("sigmaint")});

                    plotData.put(i, entry);
                */

                    Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
                }

            } catch(Exception e) {
                failed = true;
                e.printStackTrace();
            }
        }
        
        return out;
    }
    
    /**
     * Used to fit parameters front the power model, linear model, log model, or counting model.
     * @param s the model to run, given by constants in Config
     * @return a map from parameter name to a list of the parameter values for each subject
     */
    public static LinkedHashMap<String, ArrayList<String>> runModel(String s) {

        LinkedHashMap<String, ArrayList<String>> out = new LinkedHashMap<String, ArrayList<String>>();

        for (int i = 0; i < expObs.size(); i++) {
            if(!GoThread.run) {
                return null;
            }
            
            clean();

            Go.loadingWindow.addTextToLog("Estimating parameters for " + expObs.get(i).getName() + "...", true);

            boolean outlier = false;
            boolean failed = false;
            try {
                double[] exp = expObs.get(i).getExp();
                double[] obs = expObs.get(i).getObs();
                code.addDoubleArray("exp" + i, exp);
                code.addDoubleArray("obs" + i, obs);

                String call = "out" + i + " <- " + s + "(exp" + i + ",obs" + i;
                if (Config.step3Panel.outlierIsSelected()) {
                    outlier = true;
                    call = call + ", remove.outliers = TRUE, outliers.sd = " + Double.parseDouble(Config.step3Panel.outliersSD());
                }
                call = call + ");";
                code.addRCode(call);

                caller.setRCode(code);

                caller.runAndReturnResult("out" + i);
            } catch (Exception e) {
                failed = true;
                e.printStackTrace();
            }
            
            System.out.println(caller.getParser().getNames());

            if (failed || extractFromModel("logLik").equals("-Inf") || extractFromModel("logLik").equals("Inf")) {
                addToHashMap(out, "subject", expObs.get(i).getName());
                addToHashMap(out, "alpha", "NA");
                addToHashMap(out, "beta", "NA");
                addToHashMap(out, "sigma", "NA");
                addToHashMap(out, "alpha_ci-", "NA");
                addToHashMap(out, "alpha_ci+", "NA");
                addToHashMap(out, "beta_ci-", "NA");
                addToHashMap(out, "beta_ci+", "NA");
                addToHashMap(out, "sigma_ci-", "NA");
                addToHashMap(out, "sigma_ci+", "NA");
                addToHashMap(out, "num_trials", "NA");

                if (outlier) {
                    addToHashMap(out, "num_outliers", "NA");
                }

                Go.loadingWindow.addTextToLog("FAILED\nUnable to fit parameters.", LogEventType.FAILURE);

            } else {
                addToHashMap(out, "subject", expObs.get(i).getName());
                addToHashMap(out, "alpha", extractFromModel("alpha"));
                addToHashMap(out, "beta", extractFromModel("beta"));
                addToHashMap(out, "sigma", extractFromModel("sigmaint"));
                addToHashMap(out, "alpha_ci-", calculateCILower(extractFromModel("alpha"),extractFromModel("alpha_se")));
                addToHashMap(out, "alpha_ci+", calculateCIUpper(extractFromModel("alpha"),extractFromModel("alpha_se")));
                addToHashMap(out, "beta_ci-", calculateCILower(extractFromModel("beta"),extractFromModel("beta_se")));
                addToHashMap(out, "beta_ci+", calculateCIUpper(extractFromModel("beta"),extractFromModel("beta_se")));
                addToHashMap(out, "sigma_ci-", calculateCILower(extractFromModel("sigmaint"),extractFromModel("sigmaint_se")));
                addToHashMap(out, "sigma_ci+", calculateCIUpper(extractFromModel("sigmaint"),extractFromModel("sigmaint_se")));
                addToHashMap(out, "num_trials", extractFromModel("trials"));

                if (outlier) {
                    addToHashMap(out, "num_outliers", extractFromModel("num_outliers"));
                }
                
                HashMap<String, String[]> entry = plotData.get(i);
                if(entry == null) {
                    entry = new HashMap<String, String[]>();
                }

                entry.put(s,new String[]{extractFromModel("alpha"),extractFromModel("beta"),extractFromModel("sigmaint")});

                plotData.put(i, entry);
                

                Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
            }

        }

        return out;
    }

    /**
     * Runs a Wald Test with a single estimated parameter value, a test value, and the estimated parameters standard error
     * @param estValue the fitted parameter
     * @param testValue the parameter value to test against
     * @param estSE the fitted parameter's standard error
     * @return the p-value and the chi-squared value returned by a wald test
     */
    public static String[] waldTest(String estValue, double testValue, String estSE) {

        boolean failed = false;
        try {
            estValue = Double.parseDouble(estValue) + "";
            estSE = Double.parseDouble(estSE) + "";
        } catch (Exception e) {
            failed = true;
        }

        if (failed) {
            return new String[]{"NA", "NA"};
        }

        clean();
        code.addRCode("out <- " + RConfig.wald + "(" + estValue + "," + (testValue + "") + "," + estSE + ");");
        failed = false;
        try {
            caller.setRCode(code);
            caller.runAndReturnResult("out");
        } catch (Exception e) {
            failed = true;
        }

        if (failed) {
            return new String[]{"NA", "NA"};
        }

        String pValue = extractFromModel("pValue");
        
        String chi = extractFromModel("chiSq");

        try {
            double pVal = Double.parseDouble(pValue);
            pValue = Output.decimalFormat.format(pVal);
            if (Double.parseDouble(pValue) == 0) {
                pValue = pValue.substring(0, pValue.length() - 1) + "1"; //turns something like 0.0000 to 0.0001
            }
        } catch (Exception e) {
            pValue = "NA";
        }
        
        return new String[]{pValue, chi};
    }

    /**
     * Runs a Wald Test for a single subject by running individual Wald Tests for each of the test 
     * values that a user has indicated (e.g., one or more of alpha, beta, or sigma)
     * @param exp the list of intensities
     * @param obs the list of responses
     * @param model the model to fit parameters against to obtain estimated parameters
     * @return a map of parameter names to its value
     */
    public static LinkedHashMap<String, String> waldTest(double[] exp, double[] obs, String model) {
        clean();

        Step3Panel panel = Config.step3Panel;
        LinkedHashMap<String, String> out = new LinkedHashMap<String, String>();

        code.addDoubleArray("exp", exp);
        code.addDoubleArray("obs", obs);
        code.addRCode("out <- " + model + "(exp,obs);");

        boolean failed = false;
        try {
            caller.setRCode(code);
            caller.runAndReturnResult("out");
        } catch (Exception e) {
            failed = true;
        }

        String diffName = RConfig.toColloquial(model) + "_wald";
        if (!failed) {
            String alpha = extractFromModel("alpha");
            String alpha_se = extractFromModel("alpha_se");
            String beta = extractFromModel("beta");
            String beta_se = extractFromModel("beta_se");
            String sigma = extractFromModel("sigmaint");
            String sigma_se = extractFromModel("sigmaint_se");

            if (panel.alphaIsSelected()) {
                String[] s = waldTest(alpha, panel.getAlpha(), alpha_se);
                out.put(diffName + "_alpha_p", s[0]);
                out.put(diffName + "_alpha_chiSq", s[1]);
            }
            if (panel.betaIsSelected()) {
                String[] s = waldTest(beta, panel.getBeta(), beta_se);
                out.put(diffName + "_beta_p", s[0]);
                out.put(diffName + "_beta_chiSq", s[1]);
            }
            if (panel.sigmaIsSelected()) {
                String[] s = waldTest(sigma, panel.getSigma(), sigma_se);
                out.put(diffName + "_sigma_p", s[0]);
                out.put(diffName + "_sigma_chiSq", s[1]);
            }
        } else {
            if (panel.alphaIsSelected()) {
                out.put(diffName + "_alpha_p", "NA");
                out.put(diffName + "_alpha_chiSq", "NA");
            }
            if (panel.betaIsSelected()) {
                out.put(diffName + "_beta_p", "NA");
                out.put(diffName + "_beta_chiSq", "NA");
            }
            if (panel.sigmaIsSelected()) {
                out.put(diffName + "_sigma_p", "NA");
                out.put(diffName + "_sigma_chiSq", "NA");
            }
        }

        return out;
    }

    /**
     * Runs a Wald Test for all subjects and all selected models.
     * @return a map that maps from model name to a map that maps from parameter name to a list of parameter values for each subject
     */
    public static LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> waldTest() {
        LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> out = new LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>>();

        boolean power = Config.step3Panel.powerIsSelected(), linear = Config.step3Panel.linearIsSelected(), log = Config.step3Panel.logIsSelected();

        ArrayList<String> models = new ArrayList<String>();
        if (power) {
            models.add(RConfig.powerModel);
        }
        if (linear) {
            models.add(RConfig.linearModel);
        }
        if (log) {
            models.add(RConfig.logModel);
        }

        for (int i = 0; i < models.size(); i++) {
            String model = RConfig.toColloquial(models.get(i));
            LinkedHashMap<String, ArrayList<String>> singleSheet = new LinkedHashMap<String, ArrayList<String>>();

            for (int subj = 0; subj < expObs.size(); subj++) {
                if(!GoThread.run) {
                    return null;
                }
                
                Go.loadingWindow.addTextToLog("Running the Wald Test for the " + model + " model for " + expObs.get(subj).getName() + "...", true);

                LinkedHashMap<String, String> single = waldTest(expObs.get(subj).getExp(), expObs.get(subj).getObs(), models.get(i));

                Iterator<String> keys = single.keySet().iterator();
                boolean NAvalues = true;
                while (keys.hasNext()) {
                    String key = keys.next();
                    ArrayList<String> list = new ArrayList<String>();
                    if (singleSheet.containsKey(key)) {
                        list = singleSheet.get(key);
                    }
                    list.add(single.get(key));
                    singleSheet.put(key, list);

                    if (!single.get(key).equals("NA")) {
                        NAvalues = false;
                    }
                }

                if (NAvalues) {
                    Go.loadingWindow.addTextToLog("FAILED\nUnable to compute Wald Test.", LogEventType.FAILURE);
                } else {
                    Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
                }
            }

            out.put(model, singleSheet);
        }

        return out;

    }

    /**
     * Runs a likelihood ratio test for all subjects and all selected models.
     * @param alpha whether or not to use a test value for alpha
     * @param beta whether or not to use a test value for beta
     * @param sigma whether or not to use a test value for sigma
     * @return a map that maps from model name to a map that maps from parameter name to a list of parameter values for each subject
     */
    public static LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> likelihoodRatioTest(boolean alpha, boolean beta, boolean sigma) {
        LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> out = new LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>>();

        boolean power = Config.step3Panel.powerIsSelected(), linear = Config.step3Panel.linearIsSelected(), log = Config.step3Panel.logIsSelected();

        ArrayList<String> models = new ArrayList<String>();
        if (power) {
            models.add(RConfig.powerModel);
        }
        if (linear) {
            models.add(RConfig.linearModel);
        }
        if (log) {
            models.add(RConfig.logModel);
        }

        for (int i = 0; i < models.size(); i++) {
            String model = RConfig.toColloquial(models.get(i));
            LinkedHashMap<String, ArrayList<String>> singleSheet = new LinkedHashMap<String, ArrayList<String>>();

            for (int subj = 0; subj < expObs.size(); subj++) {
                if(!GoThread.run) {
                    return null;
                }
                Go.loadingWindow.addTextToLog("Finding the Likelihood Ratio for the " + model + " model for " + expObs.get(subj).getName() + "...", true);

                LinkedHashMap<String, String> single = likelihoodRatioTest(expObs.get(subj).getExp(), expObs.get(subj).getObs(), models.get(i));

                Iterator<String> keys = single.keySet().iterator();
                boolean NAvalues = true;
                while (keys.hasNext()) {
                    String key = keys.next();
                    ArrayList<String> list = new ArrayList<String>();
                    if (singleSheet.containsKey(key)) {
                        list = singleSheet.get(key);
                    }
                    list.add(single.get(key));
                    singleSheet.put(key, list);

                    if (!single.get(key).equals("NA")) {
                        NAvalues = false;
                    }
                }

                if (NAvalues) {
                    Go.loadingWindow.addTextToLog("FAILED\nUnable to compute likelihood ratio.", LogEventType.FAILURE);
                } else {
                    Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
                }
            }

            out.put(model, singleSheet);
        }

        return out;
    }

    /**
     * Runs the likelihood ratio test for a single subject and a single model
     * @param exp the list of intensities
     * @param obs the list of responses
     * @param model the model to use to obtain estimated parameter values
     * @return  a map that maps from parameter name to parameter value
     */
    public static LinkedHashMap<String, String> likelihoodRatioTest(double[] exp, double[] obs, String model) {
        clean();

        Step3Panel panel = Config.step3Panel;
        LinkedHashMap<String, String> out = new LinkedHashMap<String, String>();

        code.addDoubleArray("exp", exp);
        code.addDoubleArray("obs", obs);
        code.addRCode("out <- " + model + "(exp,obs);");

        String extra = (panel.alphaIsSelected() ? ", testAlpha=" + panel.getAlpha() : "")
                + (panel.betaIsSelected() ? ", testBeta=" + panel.getBeta() : "")
                + (panel.sigmaIsSelected() ? ", testSigma=" + panel.getSigma() : "");
        code.addRCode("llr <- " + RConfig.LikelihoodRatio + "(out" + extra + ");");

        boolean failed = false;
        try {
            caller.setRCode(code);
            caller.runAndReturnResult("llr");
        } catch (Exception e) {
            failed = true;
        }

        String diffName = RConfig.toColloquial(model) + "_likelihoodratio";
        if (failed) {
            out.put(diffName + "_p", "NA");
            out.put(diffName + "_chiSq", "NA");
        } else {
            String p = extractFromModel("pValue");
          
            
            String chi = extractFromModel("chiSq");
  
            try {
                double pVal = Double.parseDouble(p);
                p = Output.decimalFormat.format(pVal);
                if(Double.parseDouble(p) == 0) {
                    p = p.substring(0,p.length()-1) + "1"; //turns something like 0.0000 to 0.0001
                }
            } catch(Exception e) {
                p = "NA";
            }
            
            out.put(diffName + "_p", p);
            out.put(diffName + "_chiSq", chi);
        }

        return out;
    }

    
    public static LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> DIC(boolean power, boolean linear, boolean log, boolean discontinuous, boolean counting) {
        LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> out = new LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>>();

        ArrayList<String> models = new ArrayList<String>();
        if (power) {
            models.add("Power");
        }
        if (linear) {
            models.add("Linear");
        }
        if (log) {
            models.add("Log");
        }

        for (int i = 0; i < models.size(); i++) {
            String model1 = models.get(i);

            LinkedHashMap<String, ArrayList<String>> singleSheet = new LinkedHashMap<String, ArrayList<String>>();
            for (int j = 0; j < models.size(); j++) {
                if (i == j) {
                    continue;
                }

                String model2 = models.get(j);

                for (int subj = 0; subj < expObs.size(); subj++) {
                    if(!GoThread.run) {
                        return null;
                    }
                    Go.loadingWindow.addTextToLog("Finding the DIC difference between " + RConfig.toColloquial(model1) + " and " + RConfig.toColloquial(model2) + " for " + expObs.get(subj).getName() + "...", true);

                    //check if we already computed this AIC difference
                    boolean canSkip = false;
                    boolean NAvalues = true;
                    if (out.containsKey(RConfig.toColloquial(model2))) {
                        LinkedHashMap<String, ArrayList<String>> temp = out.get(RConfig.toColloquial(model2));
                        String tempKey = RConfig.toColloquial(model2) + "_" + RConfig.toColloquial(model1) + "_DICbetter";
                        if (temp.containsKey(tempKey)) {
                            ArrayList<String> allVals = temp.get(tempKey);

                            ArrayList<String> copy = new ArrayList<String>();
                            for (int f = 0; f < allVals.size(); f++) {
                                String value = allVals.get(f);
                                try {
                                    double d = Double.parseDouble(value);
                                    value = Double.toString(d);
                                } catch (Exception e) {

                                }
                                copy.add(value);
                                if (!value.equals("NA")) {
                                    NAvalues = false;
                                }
                            }

                            singleSheet.put(RConfig.toColloquial(model1) + "_" + RConfig.toColloquial(model2) + "_DICbetter", copy);
                            
                            String anotherTempKey = RConfig.toColloquial(model2) + "_" + RConfig.toColloquial(model1) + "_DICdiff";
                            ArrayList<String> anotherAllVals = temp.get(anotherTempKey);
                            ArrayList<String> anotherCopy = new ArrayList<String>();
                            for (int f = 0; f < anotherAllVals.size(); f++) {
                                String value = anotherAllVals.get(f);
                                try {
                                    double d = Double.parseDouble(value);
                                    value = Double.toString(d);
                                } catch (Exception e) {

                                }
                                anotherCopy.add(value);
                                if (!value.equals("NA")) {
                                    NAvalues = false;
                                }
                            }
                            
                            singleSheet.put(RConfig.toColloquial(model1) + "_" + RConfig.toColloquial(model2) + "_DICdiff", anotherCopy);

                            
                            canSkip = true;
                        }
                    }

                    if (!canSkip) {
                        //if we haven't, go ahead and compute it
                        LinkedHashMap<String, String> aic = DIC(expObs.get(subj).getExp(), expObs.get(subj).getObs(), model1, model2);
                        Iterator<String> params = aic.keySet().iterator();
                        while (params.hasNext()) {
                            String param = params.next();
                            ArrayList<String> all = new ArrayList<String>();
                            if (singleSheet.containsKey(param)) {
                                all = singleSheet.get(param);
                            }
                            all.add(aic.get(param));
                            if (!aic.get(param).equals("NA")) {
                                NAvalues = false;
                            }

                            singleSheet.put(param, all);
                        }
                    }

                    if (NAvalues) {
                        Go.loadingWindow.addTextToLog("FAILED\nUnable to compute DIC difference.", LogEventType.FAILURE);
                    } else {
                        Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
                    }
                }
            }

            if(model1.equals("Power"))
                model1 = "power.bayes";
            if(model1.equals("Linear"))
                model1 = "linear.bayes";
            if(model1.equals("Log"))
                model1 = "log.bayes";
            
            out.put(RConfig.toColloquial(model1), singleSheet);
        }

        return out;
    }
    
    
    /**
     * Computes the AIC differences between the various selected models for each subject
     * @param power whether or not power model is selected
     * @param linear whether or not linear model is selected
     * @param log whether or not log model is selected
     * @param discontinuous whether or not discontinuous model is selected
     * @param counting whether or not counting model is selected
     * @return a map that maps from a model's name to a map that maps from parameter name to a list of parameter values for each subject
     */
    public static LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> AIC(boolean power, boolean linear, boolean log, boolean discontinuous, boolean counting) {
        LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>> out = new LinkedHashMap<String, LinkedHashMap<String, ArrayList<String>>>();

        ArrayList<String> models = new ArrayList<String>();
        if (power) {
            models.add(RConfig.powerModel);
        }
        if (linear) {
            models.add(RConfig.linearModel);
        }
        if (log) {
            models.add(RConfig.logModel);
        }
        if (discontinuous) {
            models.add(RConfig.discontinuousModel);
        }
        if (counting) {
            models.add(RConfig.countingModel);
        }

        for (int i = 0; i < models.size(); i++) {
            String model1 = models.get(i);

            LinkedHashMap<String, ArrayList<String>> singleSheet = new LinkedHashMap<String, ArrayList<String>>();
            for (int j = 0; j < models.size(); j++) {
                if (i == j) {
                    continue;
                }

                String model2 = models.get(j);

                for (int subj = 0; subj < expObs.size(); subj++) {
                    if(!GoThread.run) {
                        return null;
                    }
                    Go.loadingWindow.addTextToLog("Finding the AIC difference between " + RConfig.toColloquial(model1) + " and " + RConfig.toColloquial(model2) + " for " + expObs.get(subj).getName() + "...", true);

                    //check if we already computed this AIC difference
                    boolean canSkip = false;
                    boolean NAvalues = true;
                    if (out.containsKey(RConfig.toColloquial(model2))) {
                        LinkedHashMap<String, ArrayList<String>> temp = out.get(RConfig.toColloquial(model2));
                        String tempKey = RConfig.toColloquial(model2) + "_" + RConfig.toColloquial(model1) + "_AICbetter";
                        if (temp.containsKey(tempKey)) {
                            ArrayList<String> allVals = temp.get(tempKey);

                            ArrayList<String> copy = new ArrayList<String>();
                            for (int f = 0; f < allVals.size(); f++) {
                                String value = allVals.get(f);
                                try {
                                    double d = Double.parseDouble(value);
                                    value = Double.toString(d);
                                } catch (Exception e) {

                                }
                                copy.add(value);
                                if (!value.equals("NA")) {
                                    NAvalues = false;
                                }
                            }

                            singleSheet.put(RConfig.toColloquial(model1) + "_" + RConfig.toColloquial(model2) + "_AICbetter", copy);
                            
                            String anotherTempKey = RConfig.toColloquial(model2) + "_" + RConfig.toColloquial(model1) + "_AICdiff";
                            ArrayList<String> anotherAllVals = temp.get(anotherTempKey);
                            ArrayList<String> anotherCopy = new ArrayList<String>();
                            for (int f = 0; f < anotherAllVals.size(); f++) {
                                String value = anotherAllVals.get(f);
                                try {
                                    double d = Double.parseDouble(value);
                                    value = Double.toString(d);
                                } catch (Exception e) {

                                }
                                anotherCopy.add(value);
                                if (!value.equals("NA")) {
                                    NAvalues = false;
                                }
                            }
                            
                            singleSheet.put(RConfig.toColloquial(model1) + "_" + RConfig.toColloquial(model2) + "_AICdiff", anotherCopy);

                            
                            canSkip = true;
                        }
                    }

                    if (!canSkip) {
                        //if we haven't, go ahead and compute it
                        LinkedHashMap<String, String> aic = AIC(expObs.get(subj).getExp(), expObs.get(subj).getObs(), model1, model2);
                        Iterator<String> params = aic.keySet().iterator();
                        while (params.hasNext()) {
                            String param = params.next();
                            ArrayList<String> all = new ArrayList<String>();
                            if (singleSheet.containsKey(param)) {
                                all = singleSheet.get(param);
                            }
                            all.add(aic.get(param));
                            if (!aic.get(param).equals("NA")) {
                                NAvalues = false;
                            }

                            singleSheet.put(param, all);
                        }
                    }

                    if (NAvalues) {
                        Go.loadingWindow.addTextToLog("FAILED\nUnable to compute AIC difference.", LogEventType.FAILURE);
                    } else {
                        Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
                    }
                }
            }

            out.put(RConfig.toColloquial(model1), singleSheet);
        }

        return out;
    }
    
    public static LinkedHashMap<String, String> DIC(double[] exp, double[] obs, String model1, String model2) {
        clean();

        LinkedHashMap<String, String> out = new LinkedHashMap<String, String>();

        
        boolean outlier = false;
        String outlierText = "remove.outliers=FALSE";
        
        code.addDoubleArray("exp", exp);
        code.addDoubleArray("obs", obs);
        

        String modelstring1 = "bayes.generateStringModel(\"" + model1 + "\"," + 
                               Config.baysPanel.getAlphaInfo() + "," + Config.baysPanel.getBetaInfo() + "," + Config.baysPanel.getSigmaInfo() + ")";
        String call1 = "bayesmodel1" + " <- bayes.estimate(model.name=\""+model1+"\",intensity=exp"+",response=obs"+",string.model="+modelstring1+",n.adapt=100, n.burn=1000, n.chains=4, n.saved=4000,"+outlierText+")";
        String final1 = "invisible(capture.output("+call1+"))";
        
        String modelstring2 = "bayes.generateStringModel(\"" + model2 + "\"," + 
                               Config.baysPanel.getAlphaInfo() + "," + Config.baysPanel.getBetaInfo() + "," + Config.baysPanel.getSigmaInfo() + ")";
        String call2 = "bayesmodel2" + " <- bayes.estimate(model.name=\""+model2+"\",intensity=exp"+",response=obs"+",string.model="+modelstring2+",n.adapt=100, n.burn=1000, n.chains=4, n.saved=4000,"+outlierText+")";
        String final2 = "invisible(capture.output("+call2+"))";
        
        code.addRCode(final1);
        code.addRCode("bayesmodel1"+"$string.model=\"deleted\"");
        code.addRCode(final2);
        code.addRCode("bayesmodel2"+"$string.model=\"deleted\"");
        

        code.addRCode("aic <- " + "bayes.DIC(dic1=bayesmodel1$model.dic, dic2=bayesmodel2$model.dic,model1.name=\""+model1+"\",model2.name=\""+model2+"\")");
        code.addRCode("aic$diffDIC = as.numeric(aic$diffDIC[1])");

        boolean failed = false;
        try {
            caller.setRCode(code);
            caller.runAndReturnResult("aic");
            
        } catch (Exception e) {

            failed = true;
        }

        System.err.println("FAILED?: " + failed);
        String diffName = RConfig.toColloquial(model1) + "_" + RConfig.toColloquial(model2);
        if (failed) {
            out.put(diffName+"_DICbetter", "NA");
            out.put(diffName+"_DICdiff", "NA");
        } else {
            out.put(diffName+"_DICbetter", extractFromModel("betterModel"));
            out.put(diffName+"_DICdiff", extractFromModel("diffDIC"));

        }
        
        return out;    
    }

    /**
     * Gets the AIC parameters between two models for a single subject given a list of expected and observed values
     * @param exp a list of intensities
     * @param obs a list of responses
     * @param model1 the first model
     * @param model2 the second model
     * @return a map from a parameter name for AIC output to parameter value
     */
    private static LinkedHashMap<String, String> AIC(double[] exp, double[] obs, String model1, String model2) {
        clean();

        LinkedHashMap<String, String> out = new LinkedHashMap<String, String>();

        code.addDoubleArray("exp", exp);
        code.addDoubleArray("obs", obs);
        code.addRCode("out1 <- " + model1 + "(exp,obs);");
        code.addRCode("out2 <- " + model2 + "(exp,obs);");
        code.addRCode("aic <- " + RConfig.AIC + "(out1,out2);");

        boolean failed = false;
        try {
            caller.setRCode(code);
            caller.runAndReturnResult("aic");
        } catch (Exception e) {

            failed = true;
        }

        String diffName = RConfig.toColloquial(model1) + "_" + RConfig.toColloquial(model2);
        if (failed) {
            out.put(diffName+"_AICbetter", "NA");
            out.put(diffName+"_AICdiff", "NA");
        } else {
            out.put(diffName+"_AICbetter", extractFromModel("betterModel"));
            out.put(diffName+"_AICdiff", extractFromModel("dAIC"));

        }

        return out;
    }

    /**
     * Adds the given key and value to a HashMap that maps from string to list of strings
     * by either creating a new entry or adding to the existing entry by inserting the value into the list
     * @param map the HashMap to add to
     * @param key the key of the entry to add
     * @param value the value of the entry to add
     */
    public static void addToHashMap(LinkedHashMap<String, ArrayList<String>> map, String key, String value) {
        if (map.containsKey((key))) {
            ArrayList<String> list = map.get(key);
            list.add(value);
            map.put(key, list);
        } else {
            ArrayList<String> list = new ArrayList<String>();
            list.add(value);
            map.put(key, list);
        }
    }

    /**
     * Using the RCaller object, extracts a particular value from the returned R result object
     * @param toExtract which parameter to extract
     * @return the value of that parameter according to R
     */
    public static String extractFromModel(String toExtract) {
        String inReturn = caller.getParser().getAsStringArray(toExtract)[0];
        if(inReturn.equals("NaN"))
            inReturn = "NA";
        return inReturn;
    }

    /**
     * Runs the counting fit for all subjects
     * @return a map from parameter name to the list of parameter values for each subject
     */
    public static LinkedHashMap<String, ArrayList<String>> countingFit() {
        return runModel(RConfig.countingModel);
    }

    /**
     * Runs the power fit for all subjects
     * @return a map from parameter name to the list of parameter values for each subject
     */
    public static LinkedHashMap<String, ArrayList<String>> powerFit() {
        return runModel(RConfig.powerModel);
    }
    
    public static LinkedHashMap<String, ArrayList<String>> bayesPowerFit() {
        return runBayesModel("Power");
    }
    
    public static LinkedHashMap<String, ArrayList<String>> bayesLinearFit() {
        return runBayesModel("Linear");
    }
    
    public static LinkedHashMap<String, ArrayList<String>> bayesLogFit() {
        return runBayesModel("Log");
    }

    /**
     * Runs the linear fit for all subjects
     * @return a map from parameter name to the list of parameter values for each subject
     */
    public static LinkedHashMap<String, ArrayList<String>> linearFit() {
        return runModel(RConfig.linearModel);
    }

    /**
     * Runs the log fit for all subjects
     * @return a map from parameter name to the list of parameter values for each subject
     */
    public static LinkedHashMap<String, ArrayList<String>> logFit() {
        return runModel(RConfig.logModel);
    }
    
    public static void generatePlots() {
        //use plot data, go through each subject, generate plots
        
        HashMap<Integer, HashMap<String, String[]>> data = R.plotData;
        for(int s : data.keySet()) {
            HashMap<String, String[]> models = data.get(s);
            
            Go.loadingWindow.addTextToLog("\nGenerating plots for " + expObs.get(s).getName()+"...", LogEventType.NEUTRAL);
            
            String[] power = models.get(RConfig.powerModel);
            String[] linear = models.get(RConfig.linearModel);
            String[] log = models.get(RConfig.logModel);
            String[] counting = models.get(RConfig.countingModel);
            String[] discon = models.get(RConfig.discontinuousModel);
            
            clean();
            code.addDoubleArray("exp", expObs.get(s).getExp());
            code.addDoubleArray("obs", expObs.get(s).getObs());
            if(power != null) {
                code.addDoubleArray("powerinfo",toDoubleArray(power));
            } else {
                code.addRCode("powerinfo=NA");
            }
            
            if(linear != null) {
                code.addDoubleArray("linearinfo",toDoubleArray(linear));
            } else {
                code.addRCode("linearinfo=NA");
            }
            
            if(log != null) {
                code.addDoubleArray("loginfo",toDoubleArray(log));
            } else {
                code.addRCode("loginfo=NA");
            }
            
            if(counting != null) {
                code.addDoubleArray("countinginfo",toDoubleArray(counting));
            } else {
                code.addRCode("countinginfo=NA");
            }
            
            if(discon != null) {
                code.addDoubleArray("disconinfo",toDoubleArray(discon));
            } else {
                code.addRCode("disconinfo=NA");
            }
            
            String outlier = "NA";
            if (Config.step3Panel.outlierIsSelected()) {
                outlier = Config.step3Panel.outliersSD();
            }
            
            //String slash = "\\";
            String call = "generatePlots(exp,obs,outlier="+outlier+",power=powerinfo,linear=linearinfo,log=loginfo,counting=countinginfo,discontinuous=disconinfo,filename=" + "\""+(Config.step4Panel.getOutputDir()+Config.step4Panel.getSavePlotImagesPrefix()+"_"+expObs.get(s).getName()).replaceAll("\\\\","\\\\\\\\")+".pdf\")";
            
            try {
                code.addRCode(call);
                caller.setRCode(code);
                caller.runOnly();
                Go.loadingWindow.addTextToLog("SUCCESS", LogEventType.SUCCESS);
            } catch(Exception e) {
                Go.loadingWindow.addTextToLog("FAILED", LogEventType.FAILURE);
            }
        }
    }

    /**
     * Reads in a text file with the provided path name and returns a large String with 
     * the contents of the file
     * @param pathname the path to the file
     * @return a String representation of that file
     */
    private static String readFile(String pathname) {
        try {
            File file = new File(pathname);
            StringBuilder fileContents = new StringBuilder((int) file.length());
            Scanner scanner = new Scanner(file);
            String lineSeparator = System.getProperty("line.separator");

            try {
                while (scanner.hasNextLine()) {
                    fileContents.append(lineSeparator + scanner.nextLine());
                }
                return fileContents.toString();
            } finally {
                scanner.close();
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }

    private static double[] toDoubleArray(String[] power) {
        double[] out = new double[power.length];
        for(int i = 0; i < power.length; i++) {
            out[i] = Double.parseDouble(power[i]);
        }
        return out;
    }

    //calculate the 95% confidence interval (lower end) given the mean and standard error
    private static String calculateCILower(String val, String val_se) {
        try {
            double v = Double.parseDouble(val);
            double v_se = Double.parseDouble(val_se);
            return Double.toString(v + 1.96*v_se);
        } catch(Exception e) {
            return "NA";
        }
    }

    //calculate the 95% confidence interval (upper end) given the mean and standard error
    private static String calculateCIUpper(String val, String val_se) {
        try {
            double v = Double.parseDouble(val);
            double v_se = Double.parseDouble(val_se);
            return Double.toString(v - 1.96*v_se);
        } catch(Exception e) {
            return "NA";
        }
    }
}
