package calculatecv;

import R.RConfig;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.PrintWriter;
import rcaller.Globals;

/**
 * Represents the Config file and operations performed on that file, including
 * reading and writing. This class also holds instances to the various panels
 * used in the GUI and the Input object used throughout to hold the user-
 * inputted data
 * 
 * @author Robert Eisinger
 */
public class Config {
    
    //The location of the RScript
    public static String RScript;
    
    //Default location for the RScript, for Unix systems
    public static String RScriptDefault = "/usr/bin/Rscript";
    
    //Location of the config file
    public static String location = "config/config";
    
    public static String dataLocation;
    
    //The data the user has inputted
    public static Input input;
    
    //The various GUI elements
    public static Step1Panel step1Panel;
    public static Step2Panel step2Panel;
    public static Step3Panel step3Panel;
    public static Step4Panel step4Panel;
    public static MainWindow mainWindow;
    public static BaysPanel baysPanel;
    
    /**
     * Initialize the config file by reading it in or creating one
     * if one doesn't exist
     */
    public static void init() {
        if(!configFileExists()) {
            createConfigFile();
        }
        
        readConfigFile();
    }
    
    /**
     * Determine if the config file exists on the user's machine in the proper
     * location
     * @return true if the config file exists, false otherwise
     */
    public static boolean configFileExists() {
        return new File(location).exists();
    }
    
    /**
     * Used to create the config file, used when first launching the application
     * or when updating config values in Step 1.
     */
    public static void createConfigFile() {
        try {
            PrintWriter out = new PrintWriter(new File(location));
            
            //save RScript location
            if(RScript != null) {
                out.println(RScript);
            } else if(!Info.isWindows()) {
                System.out.println("printing " + RScriptDefault);
                out.println(RScriptDefault);
            } else if(Info.isWindows()) {
                out.println(Globals.RScript_Windows);
            } else {
                out.println();
            }
            
            //save packagelocation, if any
            if(RConfig.packageLocation == null) {
                out.println();
            } else {
                out.println(RConfig.packageLocation);
            }
            
            //save location of data, if any
            if(Config.dataLocation == null) {
                out.println();
            } else {
                out.println(Config.dataLocation);
            }
            
            out.close();
        
        } catch (FileNotFoundException ex) {
            ex.printStackTrace();
        }
    }
    
    /**
     * Read in the config file, which is assumed to exist if this method 
     * is called, and initialize program constants, such as the location of
     * the RScript executable file.
     */
    public static void readConfigFile() {
        try {
            BufferedReader br = new BufferedReader(new FileReader(location));
            RScript = br.readLine();
            RConfig.packageLocation = br.readLine();
            Config.dataLocation = br.readLine();
            
            if(RScript.equals("")) RScript = null;
            if(RConfig.packageLocation.equals("")) RConfig.packageLocation = null;
            if(Config.dataLocation.equals("")) Config.dataLocation = null;
            
            br.close();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    /**
     * Save the config file by utilizing the createConfigFile method.
     */
    public static void save() {
        createConfigFile();
    }

    /**
     * Set the input variable by providing the path to the data files, 
     * and whether or not the files include header files. 
     * @param absolutePath the path of the data, either a file or a folder
     * @param hasHeaders whether or not the data files have headers
     */
    public static void setInput(String absolutePath,boolean hasHeaders) {
        input = new Input(absolutePath,hasHeaders);
    }
}
