package calculatecv;

import R.RConfig;
import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import rcaller.Globals;
import rcaller.RCaller;
import rcaller.RCode;

/**
 * Checks that dependencies have been satisfied, including R installation and
 * the installation of R packages Runiversal and bbmle, and provides method
 * to attempt package installation automatically.
 * 
 * @author Robert Eisinger
 */
public class DependencyChecker {

    /**
     * Validate that the RScript executable is valid.
     * 
     * @return true if the RScript executable is valid
     */
    public static boolean RScriptValidation() {
        RCaller caller = new RCaller();
        try {
            caller.setRscriptExecutable(Config.RScript);
        } catch (Exception e) {
            return false;
        }

        return true;
    }

    /**
     * Determine if a particular R package exists on the user's machine
     * @param s the package
     * @return true if the package exists
     */
    public static boolean packageExists(String s) {
        if (!RScriptValidation()) {
            return false;
        }

        try {
            RCaller caller = new RCaller();
            RCode code = new RCode();
            caller.setRscriptExecutable(Config.RScript);
            caller.cleanRCode();
            
            String x;
            if(RConfig.packageLocation == null) {
                x = "is.installed <- function(mypkg){ \n"
                         + "return(list(out=is.element(mypkg, installed.packages()[,1])))\n"
                         + "}\n"
                        + "out=is.installed(\"" + s + "\")";
            } else {
               x = "is.installed <- function(mypkg){ \n"
                         + "return(list(out=is.element(mypkg, installed.packages(lib.loc=\""+Info.getUserLocation()+"\")[,1])))\n"
                         + "}\n"
                        + "out=is.installed(\"" + s + "\")"; 
            }
            
            code.addRCode(x);
            caller.setRCode(code);

            caller.runAndReturnResult("out");
            return caller.getParser().getAsStringArray("out")[0].equals("TRUE");
        } catch (Exception e) {
            return false;
        }
    }

    /**
     * Validate that the bbmle package exists on the user's machine
     * @return true if bbmle exists
     */
    public static boolean bbmleValidation() {
        return packageExists("bbmle");
    }

    /**
     * Validate that the Runiversal package exists on this user's machine
     * @return true if Runiversal exists
     */
    public static boolean RuniversalValidation() {
        return packageExists("Runiversal");
    }

    /**
     * Attempt to install Runiversal and bbmle packages via a dynamic R
     * script that is created on the fly and uses the shell to execute
     * R. This is not guaranteed to work.
     * 
     * @param panel the GUI panel to send status messages to about this 
     * validation
     */
    public static void commandLineInstallPackages(Step1Panel panel) {
        CommandLineInstall cmi = new CommandLineInstall(panel);
        cmi.start();
    }
}

/**
 * Responsible for the command line install of R packages
 * 
 * @author Robert Eisinger
 */
class CommandLineInstall implements Runnable {

    //The Thread that the command is wrapped inside of
    private Thread thread;
    
    //The panel that the status message is sent to
    private final Step1Panel panel;

    public CommandLineInstall(Step1Panel panel) {
        this.panel = panel;
    }

    /**
     * Start the command line install process (starts the Thread run() method)
     */
    public void start() {
        thread = new Thread(this);
        thread.start();
    }

    @Override
    public void run() {
        //first try to run the default
        defaultRun();
        
        panel.checkDependencies();
        if(!Step1Panel.dependenciesSatisfied) {
            panel.pleaseWait();
            advancedRun();
            RConfig.packageLocation = Info.getUserLocation();
            panel.checkDependencies();
            if(!Step1Panel.dependenciesSatisfied) {
                RConfig.packageLocation = null; //at this point we can't help the user
            }
        }

    }

    /**
     * An advanced attempt at installing the R packages via command line, which 
     * tries to install the packages into the user's Documents folder. This is
     * used because often times, especially on PC after a fresh R install, 
     * the user may not have permissions to download R packages to the system 
     * folders that R is inside of.
     * 
     * @return true if the process didn't throw any errors
     */
    public boolean advancedRun() {
        try {

            String location = Info.getUserLocation();

            PrintWriter out = new PrintWriter(new File("depend/InstallPackagesCmdAdvanced.R"));
            out.println("has = is.element(\"Runiversal\", installed.packages()[,1]);");
            out.println("if(!has) {");
            out.println("\tinstall.packages(\"Runiversal\",repos=\""+Globals.cranRepos+"\",lib=\""+location+"\")");
            out.println("}");
            
            out.println("has = is.element(\"bbmle\", installed.packages()[,1]);");
            out.println("if(!has) {");
            out.println("\tinstall.packages(\"bbmle\",repos=\""+Globals.cranRepos+"\",lib=\""+location+"\")");
            out.println("}");
            
            out.close();
            
            
            ProcessBuilder ps = new ProcessBuilder(Config.RScript, new File(".").getAbsolutePath() + "/depend/InstallPackagesCmdAdvanced.R");

            ps.redirectErrorStream(true);

            Process pr = ps.start();

            BufferedReader in = new BufferedReader(new InputStreamReader(pr.getInputStream()));
            String line;
            while ((line = in.readLine()) != null) {
                System.out.println("INSTALL PACKAGES COMMAND ADVANCED: " + line);
            }
            pr.waitFor();

            in.close();
            
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }

        return true;
    }

    //Temp variable used for testing purposes
    boolean skipDefault = false;
    
    /**
     * Runs the default installation process which tries to install R packages
     * into the R folder in the system files. This tends to work well with Unix
     * machines but not well with PC machines, due to permission errors, in which
     * case the above method is used to install packages to the Documents folder.
     * @return true if no errors resulted
     */
    public boolean defaultRun() {
        if(skipDefault)
            return false;
        
        try {
            ProcessBuilder ps = new ProcessBuilder(Config.RScript, new File(".").getAbsolutePath() + "/depend/InstallPackagesCmd.R");
            ps.redirectErrorStream(true);

            Process pr = ps.start();

            BufferedReader in = new BufferedReader(new InputStreamReader(pr.getInputStream()));
            String line;
            while ((line = in.readLine()) != null) {
                System.out.println("INSTALL PACKAGES COMMAND: " + line);
            }
            pr.waitFor();

            in.close();


        } catch (Exception ex) {
            return false;
        }

        return true;
    }

}
