package calculatecv;

import java.util.ArrayList;

/**
 * An ExpObs object contains the expected (intensity) and observed (responses)
 * values for a single psychophysical experiment for a single subject, including
 * the subject's file name and their intensity and responses held as decimal arrays.
 * 
 * @author Robert Eisinger
 */
public class ExpObs {

    //The file name of the subject
    private final String name;
    
    //Expected and observed arrays
    private final double[] exp, obs;

    /**
     * Create an ExpObs object from a file name, a list of expected values, and
     * a list of observed values, which are assumed to be proper values (i.e., 
     * decimal values).
     * 
     * @param name the name of the subject's data file
     * @param exp list of expected decimal values as Strings
     * @param obs list of observed decimal values as Strings
     */
    public ExpObs(String name, ArrayList<String> exp, ArrayList<String> obs) {
        this.name = name;
        this.exp = asDoubleArray(exp);
        this.obs = asDoubleArray(obs);
    }

    /**
     * Get the expected values
     * @return the expected values
     */
    public double[] getExp() {
        return exp;
    }

    /**
     * Get the observed values
     * @return the observed values
     */
    public double[] getObs() {
        return obs;
    }
    
    /**
     * Convert an ArrayList<String> to double[] 
     * @param list the list of String values
     * @return an array of decimal values
     */
    private static double[] asDoubleArray(ArrayList<String> list) {
        double[] out = new double[list.size()];
        for(int i = 0; i < out.length; i++) {
            out[i] = Double.parseDouble(list.get(i));
        }
        return out;
    }

    /**
     * Get the subject's file name
     * @return the subject's file name
     */
    public String getName() {
        return name;
    }
}
