package calculatecv;

import java.awt.Color;
import javax.swing.text.BadLocationException;
import javax.swing.text.Style;
import javax.swing.text.StyleConstants;
import javax.swing.text.StyleContext;
import javax.swing.text.StyledDocument;

/**
 * The GoLoadingWindow is used to display the progress of the R code that is
 * being run behind scenes. It includes a loading bar as well as detailed output 
 * messages.
 * 
 * @author Robert Eisinger
 */
public class GoLoadingWindow extends javax.swing.JFrame {

    //Styles used for the message output
    private static Style regular, italic, bold, red, green;
    
    /**
     * Creates a GoLoadingWindow and launches the frame
     */
    public GoLoadingWindow() {
        initComponents();
        int x = Config.mainWindow.getLocation().x + (Config.mainWindow.getSize().width/2) - (getWidth()/2);
        int y = Config.mainWindow.getLocation().y + (Config.mainWindow.getSize().height/2) - (getHeight()/2);
        Config.mainWindow.setVisible(false);

        setLocation(x,y);
        setVisible(true);
        
        StyledDocument doc = log.getStyledDocument();

        // Load the default style and add it as the "regular" text
        Style def = StyleContext.getDefaultStyleContext().getStyle( StyleContext.DEFAULT_STYLE );
        regular = doc.addStyle("regular", def );

        // Create an italic style
        italic = doc.addStyle( "italic", regular );
        StyleConstants.setItalic( italic, true );

        // Create a bold style
        bold = doc.addStyle("bold", regular );
        StyleConstants.setBold( bold, true );

        // Create a red style
        red = doc.addStyle("red", regular);
        StyleConstants.setForeground(red, Color.red);
        
        // Create a green style
        green = doc.addStyle("green", regular);
        StyleConstants.setForeground(green, new Color(0, 100, 0));
    }
    
    /**
     * An error occurred - displays the error message.
     * @param error the error message
     */
    public void error(String error) {
        loadingBar.setIndeterminate(false);
        loadingBar.setValue(100);
        statusLabel.setText("Errors occurred. Please see the log below.");
        titleLabel.setText("Error");
    }
    
    /**
     * Execution finished and was successful - display the success message
     */
    public void success() {
        loadingBar.setIndeterminate(false);
        loadingBar.setValue(100);
        statusLabel.setText("Success!");
        titleLabel.setText("Complete");
        
        this.addTextToLog("\n\nRunning is complete.",true,LogEventType.SUCCESS);
    }
    
    /**
     * Adds text to the message log
     * @param text the text to add
     * @param type what type of log event is this
     */
    public void addTextToLog(String text,LogEventType type) {
        addTextToLog(text,false,type);
    }
    
    /**
     * Adds text to the message log and defaults to LogEventType.NEUTRAL (black
     * text)
     * @param text the text to add
     */
    public void addTextToLog(String text) {
        addTextToLog(text,false,LogEventType.NEUTRAL);
    }
    
    /**
     * Adds text to the log and defaults to LogEventType.NEUTRAL (black text)
     * @param text the text to add
     * @param addNewLine whether or not to append a new line character to the end
     */
    public void addTextToLog(String text, boolean addNewLine) {
        addTextToLog(text,addNewLine,LogEventType.NEUTRAL);
    }
    
    /**
     * Adds text to the log
     * @param text the text to add
     * @param addNewLineCharacter whether or not to append a new line character
     * @param type the LogEventType of the text
     */
    public void addTextToLog(String text, boolean addNewLineCharacter,LogEventType type) {
        if(text.indexOf("\n") != 0 && addNewLineCharacter)
            text = "\n" + text;
        
        Style style = regular;
        if(type == LogEventType.SUCCESS)
            style = green;
        else if(type == LogEventType.FAILURE)
            style = red;
       
        
        try {
            log.getStyledDocument().insertString(log.getStyledDocument().getLength(), text, style);
        } catch (BadLocationException ex) {
            ex.printStackTrace();
        }
        
        log.setCaretPosition(log.getDocument().getLength());

    }

    /**
     * This method is called from within the constructor to initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is always
     * regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        loadingBar = new javax.swing.JProgressBar();
        titleLabel = new javax.swing.JLabel();
        jButton1 = new javax.swing.JButton();
        statusLabel = new javax.swing.JLabel();
        jScrollPane3 = new javax.swing.JScrollPane();
        log = new javax.swing.JTextPane();

        addWindowListener(new java.awt.event.WindowAdapter() {
            public void windowClosing(java.awt.event.WindowEvent evt) {
                formWindowClosing(evt);
            }
            public void windowClosed(java.awt.event.WindowEvent evt) {
                formWindowClosed(evt);
            }
        });

        loadingBar.setIndeterminate(true);

        titleLabel.setFont(new java.awt.Font("Lucida Grande", 1, 18)); // NOI18N
        titleLabel.setHorizontalAlignment(javax.swing.SwingConstants.CENTER);
        titleLabel.setText("Loading");

        jButton1.setText("Close");
        jButton1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                jButton1ActionPerformed(evt);
            }
        });

        statusLabel.setText("Please wait...");

        log.setEditable(false);
        jScrollPane3.setViewportView(log);

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(getContentPane());
        getContentPane().setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(titleLabel, javax.swing.GroupLayout.Alignment.TRAILING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                    .addComponent(jScrollPane3)
                    .addComponent(loadingBar, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 540, Short.MAX_VALUE)
                    .addComponent(statusLabel, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
                .addContainerGap())
            .addGroup(layout.createSequentialGroup()
                .addGap(0, 0, Short.MAX_VALUE)
                .addComponent(jButton1))
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                .addComponent(titleLabel)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addComponent(statusLabel)
                .addGap(7, 7, 7)
                .addComponent(loadingBar, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addComponent(jScrollPane3, javax.swing.GroupLayout.DEFAULT_SIZE, 157, Short.MAX_VALUE)
                .addGap(18, 18, 18)
                .addComponent(jButton1))
        );

        pack();
    }// </editor-fold>//GEN-END:initComponents

    private void jButton1ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jButton1ActionPerformed
        close();
    }//GEN-LAST:event_jButton1ActionPerformed

    private void formWindowClosed(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosed
    }//GEN-LAST:event_formWindowClosed

    private void formWindowClosing(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_formWindowClosing
        close();
    }//GEN-LAST:event_formWindowClosing

    /**
     * Close the window and halt running.
     */
    public void close() {
        GoThread.run = false;
        setVisible(false);
        Config.mainWindow.setVisible(true);
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton jButton1;
    private javax.swing.JScrollPane jScrollPane3;
    private javax.swing.JProgressBar loadingBar;
    private javax.swing.JTextPane log;
    private javax.swing.JLabel statusLabel;
    private javax.swing.JLabel titleLabel;
    // End of variables declaration//GEN-END:variables

}
